from kooki.tools import get_front_matter, DictAsMember
import em

def Caller(file_full_path, metadata={}):
    def call(*args, **kwargs):
        with open(file_full_path,'r') as f:
            file_content = f.read()
            interpreter = em.Interpreter()
            interpreter.setPrefix('@')
            front_matter, new_content = get_front_matter(file_content)
            metadata_copy = metadata.copy()
            metadata_copy.update(front_matter)
            metadata_copy.update({**kwargs})
            new_metadata = DictAsMember.convert(metadata_copy)
            try:
                ret_content = interpreter.expand(new_content, new_metadata)
            except:
                ret_content = new_content
            return ret_content
    return call



import yaml
from kooki.tools import read_file
from kooki.jars import get_jars
from .command import Command

from vcstool.commands.import_ import generate_jobs as import_generate_jobs
from vcstool.commands.import_ import add_dependencies as import_add_dependencies
from vcstool.commands.import_ import get_repositories as import_get_repositories

__command__ = 'update'
__description__ = 'Update a kooki project'


class UpdateCommand(Command):

    def __init__(self):
        super(UpdateCommand, self).__init__(__command__, __description__)

    def callback(self, args):
        freeze_file = read_file('.kooki_freeze')
        freeze_data = yaml.safe_load(freeze_file)
        apply_kooki_freeze(freeze_data, freeze_file)


def apply_kooki_freeze(freeze_data, freeze_file):
    jars = get_jars()
    Output.start_step('jars')
    for jar in freeze_data['repositories']:
        version_freezed = freeze_data['repositories'][jar]['version']
        Output._print_colored(jar, 'blue')
        Output._print_colored('  freezed: ', 'yellow', end='')
        Output._print_colored(version_freezed, 'cyan')
        Output._print_colored('  installed: ', 'yellow', end='')
        if jar in jars:
            version_installed = jars[jar]['version']
            if version_freezed == version_installed:
                Output._print_colored(version_installed, 'green')
            else:
                Output._print_colored('Installing new version')
                apply_vcs_import(freeze_file)
        else:
            Output._print_colored('Installing new jar')
            apply_vcs_import(freeze_file)


def apply_vcs_import(freeze_file):
    resources_dir = get_kooki_dir()
    user_jars_dir = os.path.join(resources_dir, 'jars')
    repos = import_get_repositories(freeze_file)
    import_args = argparse.Namespace()
    import_args.path = user_jars_dir
    import_args.force = False

    jobs = import_generate_jobs(repos, import_args)
    import_add_dependencies(jobs)

    results = execute_jobs(jobs, show_progress=True, number_of_workers=1, debug_jobs=False)
    output_results(results)



import argparse, traceback
import pretty_output

from kooki.exception import KookiException

class Command:

    def __init__(self, command, description):
        self.command = command

        self.parser = argparse.ArgumentParser(prog=self.command, description=description, add_help=False)
        self.parser.set_defaults(callback=self)

        self.parser.add_argument('-d', '--debug', help='Show information to help debug the bake processing', action='store_true')
        self.parser.add_argument('--no-color', help='The output has no color.', action='store_true')
        self.parser.add_argument('--no-output', help='There is no output.', action='store_true')
        self.subparsers = None

    def add_command(self, command, description='', help=''):
        if self.subparsers == None:
            self.subparsers = self.parser.add_subparsers(help=help)
        subparser = self.subparsers.add_parser(command.parser.prog, help=command.parser.description)
        subparser.__dict__ = command.parser.__dict__;

    def add_argument(self, *args, **kwargs):
        self.parser.add_argument(*args, **kwargs)

    def run(self):
        args = self.parser.parse_args()
        args.callback(args)

    def __call__(self, args):
        try:
            pretty_output.set_output_policy(not args.no_output)
            pretty_output.set_color_policy(not args.no_color)
            pretty_output.set_debug_policy(args.debug)
            pretty_output.command(self.command)
            self.callback(args)

        except KookiException as e:
            pretty_output.error_step('Errors')
            pretty_output.error(e)

        except Exception as e:
            pretty_output.error_step('Errors')
            pretty_output.error(traceback.format_exc()[:-1])

        finally:
            pretty_output.command()



from kooki.jars import search_file
from kooki.tools import get_front_matter, DictAsMember
import em
import pretty_output

def Reader(document, metadata={}):
    def read(file_path):
        file_full_path = search_file(document.jars, document.recipe, file_path)
        ret_content = ''

        output = []
        if file_full_path:
            output.append({'name': file_path, 'status': '[found]', 'path': file_full_path})

            with open(file_full_path,'r') as f:
                file_content = f.read()
                interpreter = em.Interpreter()
                interpreter.setPrefix('@')
                front_matter, new_content = get_front_matter(file_content)
                metadata_copy = metadata.copy()
                metadata_copy.update(front_matter)
                new_metadata = DictAsMember.convert(metadata_copy)
                try:
                    ret_content = interpreter.expand(new_content, new_metadata)
                except:
                    ret_content = new_content
        else:
            output.append({'name': file_path, 'status': ('[missing]', 'red'), 'path': ''})

        pretty_output.infos(output, [('name', 'blue'), ('status', 'green'), ('path', 'cyan')])

        return ret_content

    return read



from kooki.tools import write_file
import pretty_output

def Export(name, content, extension):
    pretty_output.start_step('export')
    file_name = '{0}.{1}'.format(name, extension)
    write_file(file_name, content)
    infos = []
    infos.append({'name': file_name, 'status': '[created]', 'info': ''})
    pretty_output.infos(infos, [('name', 'blue'), ('status', 'green'), ('info', 'cyan')])



from kooki.tools import get_extension
import yaml
import toml
import json
import os

def Metadata(document):
    metadata_dict = {}
    for file_path, file_content in document.metadata.items():
        metadata = load(file_path, file_content)
        metadata_dict = data_merge(metadata, metadata_dict)
    return metadata_dict


def load(file_path, file_content):

    extension = get_extension(file_path)

    if extension == '':
        metadata = load_metadata(file_content)
    else:
        if extension == 'yaml' or extension == 'yml':
            metadata = yaml.safe_load(file_content)
        elif extension == 'toml':
            metadata = toml.loads(file_content)
        elif extension == 'json':
            metadata = json.loads(file_content)

    return metadata


def load_metadata(file_path, file_content):

    metadata_loaders = []
    metadata_loaders.append(yaml.safe_load)
    metadata_loaders.append(toml.loads)
    metadata_loaders.append(json.loads)

    for loader in metadata_loaders:
        try:
            metadata = loader(file_content)
            return metadata
        except:
            pass

    raise(Exception('Cannot load metadata provided: {}'.format(file_path)))


def data_merge(a, b):

    class MergeError(Exception):
        pass

    key = None

    try:
        if a is None or isinstance(a, str) or isinstance(a, bytes) or isinstance(a, int) or isinstance(a, float):
            a = b
        elif isinstance(a, list):
            if isinstance(b, list):
                a.extend(b)
            else:
                a.append(b)
        elif isinstance(a, dict):
            if isinstance(b, dict):
                for key in b:
                    if key in a:
                        a[key] = data_merge(a[key], b[key])
                    else:
                        a[key] = b[key]
            else:
                raise MergeError('Cannot merge non-dict "%s" into dict "%s"' % (b, a))
        else:
            raise MergeError('NOT IMPLEMENTED "%s" into "%s"' % (b, a))
    except TypeError as e:
        raise MergeError('TypeError "%s" in key "%s" when merging "%s" into "%s"' % (e, key, b, a))
    return a



from kooki.tools import get_front_matter, DictAsMember
import em

def Content(content, metadata):
    interpreter = em.Interpreter()
    interpreter.setPrefix('@')
    ret_content = ''
    for file_path, file_content in content.items():
        front_matter, new_content = get_front_matter(file_content)
        metadata_copy = front_matter.copy()
        metadata_copy.update(metadata)
        metadata = DictAsMember.convert(metadata_copy)
        ret_content += interpreter.expand(new_content, metadata)
    return ret_content



import yaml, os, argparse
import pretty_output

from kooki.config import read_config_file, get_kooki_dir_jars, get_kooki_dir_recipes
from kooki.tools import write_file, read_file
from kooki.jars import get_jars
from kooki.rule_parser import parse_document_rules

from .command import Command

__command__ = 'freeze'
__description__ = 'Freeze version of Kooki and Jars'


class FreezeCommand(Command):

    def __init__(self):
        super(FreezeCommand, self).__init__(__command__, __description__)
        self.add_argument('documents', nargs='*')
        self.add_argument('-f', '--config-file', default='kooki.yaml')

    def callback(self, args):
        config = read_config_file(args.config_file)
        document_rules = parse_document_rules(config)
        version = freeze_kooki()
        jars = freeze_jars(document_rules)
        recipes = freeze_recipes(document_rules)

        repositories = {
            'kooki': version,
            'repositories': jars}

        write_file('.kooki_freeze', yaml.safe_dump(repositories, default_flow_style=False))


def freeze_kooki():
    from kooki.version import __version__
    pretty_output.start_step('kooki')
    pretty_output._print_colored('version ', 'blue', end='')
    pretty_output._print_colored(__version__, 'cyan')
    return __version__


def freeze_jars(documents):
    from vcstool.commands.export import ExportCommand
    from vcstool.crawler import find_repositories
    from vcstool.executor import execute_jobs, generate_jobs

    pretty_output.start_step('jars')

    infos = []
    jars = set()
    export_jars = {}

    user_jars_dir = get_kooki_dir_jars()

    for name, document in documents.items():
        for jar in document.jars:
            jars.add(jar)

    for jar in jars:
        jar_path = os.path.join(user_jars_dir, jar)

        pretty_output._print_colored(jar, 'blue', end='')
        pretty_output._print_colored(' [found] ', 'green', end='')
        pretty_output._print_colored(jar_path, 'cyan')

        if os.path.isdir(jar_path):
            export_args = argparse.Namespace()
            export_args.path = jar_path
            export_args.exact = True
            command = ExportCommand(export_args)
            clients = find_repositories([jar_path])
            jobs = generate_jobs(clients, command)
            results = execute_jobs(jobs)

            pretty_output._print_colored('  vcs: ', 'yellow', end='')
            if results == []:
                pretty_output._print_colored('no vcs set', 'red')
            elif len(results) == 1:
                result = results[0]
                export_data = result['export_data']
                vcs_type = result['client'].__class__.type
                vcs_url = export_data['url']
                vcs_version = export_data['version']

                pretty_output._print_colored(vcs_type, 'cyan')
                pretty_output._print_colored('  url: ', 'yellow', end='')
                pretty_output._print_colored(vcs_url, 'cyan')
                pretty_output._print_colored('  version: ', 'yellow', end='')
                pretty_output._print_colored(vcs_version, 'cyan')

                save_jar = {}
                save_jar['type'] = vcs_type
                save_jar['url'] = vcs_url
                save_jar['version'] = vcs_version
                export_jars[jar] = save_jar

        else:
            pretty_output._print_colored('[missing]', 'red')

    return export_jars

def freeze_recipes(documents):
    from vcstool.commands.export import ExportCommand
    from vcstool.crawler import find_repositories
    from vcstool.executor import execute_jobs, generate_jobs

    pretty_output.start_step('recipes')

    recipes = set()
    export_recipes = {}

    user_recipes_dir = get_kooki_dir_recipes()

    for name, document in documents.items():
        recipes.add(document.recipe)

    for recipe in recipes:
        recipe_path = os.path.join(user_recipes_dir, recipe)

        pretty_output._print_colored(recipe, 'blue', end='')
        pretty_output._print_colored(' [found] ', 'green', end='')
        pretty_output._print_colored(recipe_path, 'cyan')

        if os.path.isdir(recipe_path):
            export_args = argparse.Namespace()
            export_args.path = recipe_path
            export_args.exact = True
            command = ExportCommand(export_args)
            clients = find_repositories([recipe_path])
            jobs = generate_jobs(clients, command)
            results = execute_jobs(jobs)

            pretty_output._print_colored('  vcs: ', 'yellow', end='')
            if results == []:
                pretty_output._print_colored('no vcs set', 'red')
            elif len(results) == 1:
                result = results[0]
                export_data = result['export_data']
                vcs_type = result['client'].__class__.type
                vcs_url = export_data['url']
                vcs_version = export_data['version']

                pretty_output._print_colored(vcs_type, 'cyan')
                pretty_output._print_colored('  url: ', 'yellow', end='')
                pretty_output._print_colored(vcs_url, 'cyan')
                pretty_output._print_colored('  version: ', 'yellow', end='')
                pretty_output._print_colored(vcs_version, 'cyan')

                export_recipe = {}
                export_recipe['type'] = vcs_type
                export_recipe['url'] = vcs_url
                export_recipe['version'] = vcs_version
                export_recipes[recipe] = export_recipe

        else:
            pretty_output._print_colored('[missing]', 'red')

    return export_recipes



from .bake import BakeCommand
from .check import CheckCommand
from .command import Command
from .freeze import FreezeCommand
from .jar import JarCommand
from .new import NewCommand
from .update import UpdateCommand



from html.parser import HTMLParser


def HTMLToTeX(content):
    parser = HTMLTeXParser()
    tex_content = parser.convert(content)
    return tex_content


class HTMLTableTeXParser(HTMLParser):

    def __init__(self):
        HTMLParser.__init__(self)
        self.clean()

    def clean(self):
        self._content = ''
        self._last_tag = ''

    def convert(self, content):
        self._table = []
        self._line = []
        self.feed(content)

        col_formatting = ''
        if len(self._table) > 0:
            nb_col = len(self._table[0])
            for i in range(0, nb_col-1):
                col_formatting += 'l '
            col_formatting += 'l'

        core = ''
        if len(self._table) > 0:

            for element in self._table[0][:-1]:
                core += '\\textbf{' + element + '} & '
            core += '\\textbf{' + self._table[0][-1] + '}'
            core += ' \\\\  \n'
            core += '\hline \n'

            for line in self._table[1:-1]:
                if len(line) > 0:
                    for element in line[:-1]:
                        core += element + ' & '
                    core += line[-1]
                core += ' \\\\ \n'

            for element in self._table[-1][:-1]:
                core += element + ' & '
            core += self._table[-1][-1]
            core += ' \\\\'

        content = '''
\\def\\arraystretch{1.5}
\\begin{center}
{
\\begin{tabular}{%s}
\hline
%s
\hline
\\end{tabular}
}
\\end{center}
''' % (col_formatting, core)

        return content

    def handle_starttag(self, tag, attrs):
        self._last_tag = tag

    def handle_endtag(self, tag):

        if tag == 'tr':
            self._table.append(self._line)
            self._line = []

    def handle_data(self, data):

        if data != '\n':

            if self._last_tag == 'td':
                self._line.append(data)

            elif self._last_tag == 'th':
                self._line.append(data)

class HTMLTeXParser(HTMLParser):

    start_tag = {
        'h1': '\\section{',
        'h2': '\\subsection{',
        'h3': '\\subsubsection{',
        'h4': '\\paragraph{',
        'h5': '\\paragraph{',
        'h6': '\\subparagraph{',
        'p': '',
        'ul': '\\begin{itemize}',
        'ol': '\\begin{enumerate}',
        'pre': '\\begin{verbatim}\n',
        'li': '\item ',
        'span': '',
        'div': '',
        'code': '\\texttt{',
        'em': '\\emph{',
        'strong': '\\textbf{',
        'dl': '\\begin{description}',
        'dt': '\item [',
        'dd': '',
        'hr': '{\noindent}\hrulefill',
    }

    end_tag = {
        'h1': '}',
        'h2': '}',
        'h3': '}',
        'h4': '}',
        'h5': '}',
        'h6': '}',
        'p': '\n',
        'ul': '\\end{itemize}',
        'ol': '\\end{enumerate}',
        'pre': '\\end{verbatim}\n',
        'li': '',
        'span': '',
        'div': '',
        'code': '}',
        'em': '}',
        'strong': '}',
        'dl': '\\end{description}',
        'dt': ']',
        'dd': '',
        'hr': '',
    }

    def __init__(self):
        HTMLParser.__init__(self)
        self.clean()

    def clean(self):
        self._content = ''
        self._last_opened = []
        self._pre_opened = False
        self._table = ''
        self._table_opened = False
        self._table_parser = HTMLTableTeXParser()
        self._table_parser.clean()
        self._link_opened = False
        self._link = ''

    def convert(self, content):
        self.feed(content)
        return self._content

    def handle_starttag(self, tag, attrs):

        self._last_opened.append(tag)

        if tag == 'pre':
            self._pre_opened = True

        if tag == 'code' and len(self._last_opened) >= 2 and self._last_opened[-2] == 'pre':
            pass

        elif tag == 'table':
            self._table_opened = True
            attrs_formatted = ' '
            for attr in attrs:
                attrs_formatted += '{0}="{1}"'.format(attr[0], attr[1])

            self._table += '<{0}{1}>'.format(tag, attrs_formatted)

        elif tag == 'a':
            self._link_opened = True
            link = attrs[0][1]
            self._link = link
            self._content += '\\href{' + link + '}{'

        elif tag in HTMLTeXParser.start_tag:
            self._content += HTMLTeXParser.start_tag[tag]

        else:
            attrs_formatted = ' '
            for attr in attrs:
                attrs_formatted += '{0}="{1}"'.format(attr[0], attr[1])

            if self._table_opened:
                self._table += '<{0}{1}>'.format(tag, attrs_formatted)
            else:
                self._content += '<{0}{1}>'.format(tag, attrs_formatted)

    def handle_endtag(self, tag):

        if tag == 'pre':
            self._pre_opened = False

        if tag == 'code' and len(self._last_opened) >= 2 and self._last_opened[-2] == 'pre':
            pass

        elif tag == 'table':
            self._table_opened = False
            self._table += '</{0}>'.format(tag)

            self._content += self._table_parser.convert(self._table)

            self._table = ''

        elif tag == 'a':
            self._link_opened = False
            self._content += '}'

        elif tag in HTMLTeXParser.end_tag:
            self._content += HTMLTeXParser.end_tag[tag]

        else:

            if self._table_opened:
                self._table += '</{0}>'.format(tag)
            else:
                self._content += '</{0}>'.format(tag)

        self._last_opened.pop()

    def handle_data(self, data):

        if not self._pre_opened:
            data = data.replace('_', '\_')

            data = data.replace('É', "\\'E")
            data = data.replace('é', "\\'e")
            data = data.replace('È', '\`e')
            data = data.replace('è', '\`e')
            data = data.replace('Ê', '\^e')
            data = data.replace('ê', '\^e')

            data = data.replace('À', '\`A')
            data = data.replace('à', '\`a')
            data = data.replace('Â', '\^A')
            data = data.replace('â', '\^a')

        if self._table_opened:
            self._table += data
        else:
            self._content += data



import os

def Extension(jars, recipe):
    current = os.getcwd()
    extensions = {}
    path = os.path.join(current, 'extensions')
    if os.path.isdir(path):
        load_extensions(path, extensions)
    for jar in jars:
        path = os.path.join(jar, recipe, 'extensions')
        if os.path.isdir(path):
            load_extensions(path, extensions)
        path = os.path.join(jar, 'extensions')
        if os.path.isdir(path):
            load_extensions(path, extensions)
    return extensions


def load_extensions(directory, extensions):
    for file_name in os.listdir(directory):
        path = os.path.join(directory, file_name)
        extension_name = os.path.splitext(file_name)[0]
        if not extension_name in extensions:
            extensions[extension_name] = path



import markdown

def MarkdownToHTML(content):

    extensions = [
        'markdown.extensions.meta',
        'markdown.extensions.tables',
        'markdown.extensions.attr_list',
        'markdown.extensions.fenced_code',
        'markdown.extensions.def_list',
        'markdown.extensions.sane_lists',
        'markdown.extensions.codehilite']

    md = markdown.Markdown(extensions=extensions)

    html_content = md.convert(content)
    return html_content



from .content import Content
from .metadata import Metadata
from .xelatex import Xelatex
from .template import Template
from .export import Export
from .split import Split
from .html_to_tex import *
from .markdown_to_html import *
from .reader import *
from .extension import Extension
from .caller import Caller
from .merge import Merge
from .custom import Custom



import em
from kooki.tools import DictAsMember

def Template(data, metadata):
    interpreter = em.Interpreter()
    interpreter.setPrefix('@')
    metadata_member = DictAsMember.convert(metadata)

    if type(data) == list:
        result = []
        for d in data:
            result.append(interpreter.expand(d, metadata_member))
    else:
        result = interpreter.expand(data, metadata_member)

    return result



import pretty_output

from kooki.jars import get_jars
from .command import Command

__command__ = 'jar'
__description__ = 'List jars'

class JarCommand(Command):

    def __init__(self):
        super(JarCommand, self).__init__(__command__, __description__)

    def callback(self, args):
        pass
